#!/usr/bin/env python

# Copyright (c) 2008 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

import os
import sys
import re
import types
import pprint
import logging

# Komodo modules.
import process
from xpcom import ServerException, nsError

log = logging.getLogger('gitlib')
#log.setLevel(logging.DEBUG)

#---- exceptions

class GITLibError(ServerException):
    def __init__(self, *args, **kwargs):
        for key in ("error", "cwd"):
            if kwargs.get(key, None) is not None:
                setattr(self, key, kwargs[key])
                del kwargs[key]
        log.debug(kwargs.items())
        ServerException.__init__(self, nsError.NS_ERROR_FAILURE, *args)
    def __getitem__(self, key):
        return getattr(self, key)
    def __contains__(self, key):
        return hasattr(self, key)
    def get(self, key, default=None):
        if key in self:
            return self[key]
        return default

class GIT:
    """A proxy to the Git command line app."""
    def __init__(self, executable='git', **options):
        self.executable = executable
        self.optd = options
        # Abortable process handle.
        self._processHelper = process.AbortableProcessHelper()

    def _runCommand(self, argv, cwd=None, env=None, input=None,
                    ignore_exit_status=False, terminalHandler=None):
        """Prepare and run the given arg vector, 'argv', and return the
        results.  Returns (<stdout>, <stderr>).
        """
        if type(argv) not in (types.ListType, types.TupleType):
            raise GITLibError("_runCommand:: argv is not a list: %r" % (argv, ),
                              cwd=cwd)
        # Remove any empty options from the cmd list.
        argv = [ x for x in argv if x ]
        log.debug("Running %r in cwd %r", argv, cwd)
        output = None
        if env is None:
            import koprocessutils
            env = koprocessutils.getUserEnv()

        # Komodo can only handle svn messages in english.
        # http://bugs.activestate.com/show_bug.cgi?id=45677
        env['LC_MESSAGES'] = 'en_US'
        # Set LANGUAGE too, otherwise it may still come back in another lang
        # http://bugs.activestate.com/show_bug.cgi?id=68615
        env['LANGUAGE'] = 'en_US'

        # A light wrapper around process.ProcessOpen.
        p = self._processHelper.ProcessOpen(cmd=argv, cwd=cwd, env=env,
                                            universal_newlines=True)
        try:
            if input:
                input = input.encode("utf-8")
            if terminalHandler:
                terminalHandler.hookIO(p.stdin, p.stdout, p.stderr, " ".join(argv))
                p.wait()
                # Output and errors have gone to the terminal.
                output = ''
                error = ''
            else:
                output, error = p.communicate(input)
            retval = p.returncode
            p.close()
        finally:
            self._processHelper.ProcessDone()

        if retval != 0 and not ignore_exit_status:
            # If the error message _looks_ like a fatal error, use it. Otherwise
            # prefer having no details over being possibly wrong.
            shorterror = ''.join(error)
            if not shorterror.startswith("fatal:"):
                shorterror = None
            raise GITLibError("Error %s running '%s'\n  in '%s': \n%s"\
                              % (retval, argv, cwd, ''.join(error)),
                              error=shorterror,
                              cwd=cwd)

        return output, error

    def abort(self):
        self._processHelper.ProcessAbort()

    def add(self, files, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "add"]
        if options:
            argv += options.split(" ")
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    def clone(self, urls, cwd=None, altdirname=None, command_options="",
              terminalHandler=None, env=None):
        if type(urls) in types.StringTypes:
            urls = [urls]
        argv = [self.executable, 'clone']
        if command_options:
            # Example: "--rev 103 --clean"
            argv += command_options.split(" ")
        argv += urls
        if altdirname:
            argv.append(altdirname)
        output, error = self._runCommand(argv,
                                         cwd=cwd,
                                         env=env,
                                         terminalHandler=terminalHandler)
        result = {}
        result['stdout'] = output
        result['stderr'] = error
        return result

    def commit(self, files, cwd=None, options=None, env=None, msgpath=None):
        result = {}
        argv = [self.executable, "commit"]
        if options:
            argv += options.split(" ")
        if msgpath is not None:
            argv += [ "-F", msgpath ]
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    def diff(self, files, cwd=None, options=None, external=False, env=None):
        result = {}
        if external:
            argv = [self.executable, "difftool"]
        else:
            argv = [self.executable, "diff"]
        if options:
            argv += options.split(" ")
        if files:
            argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    log_revisions_re = re.compile(
        r'''commit\s+(?P<revision>\w*?)\n''' \
         '''Author:\s+(?P<author>.*?)\n''' \
         '''Date:\s+(?P<date>.*?)\n''' \
         '''\n'''
    )

    def _formatLogMessage(self, message):
        result = []
        for line in message.rstrip().splitlines(0):
            if line.startswith("    "):
                result.append(line[4:])
            else:
                result.append(line)
        return "\n".join(result) + "\n"

    def log(self, path, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "log"]
        if options:
            argv += options.split(" ")
        argv.append(path)
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error

        log_items = []
        matchIterator = self.log_revisions_re.finditer(output)
        try:
            match = matchIterator.next()
        except StopIteration:
            match = None
        while match:
            text_start = match.end()
            try:
                matchNext = matchIterator.next()
            except StopIteration:
                matchNext = None
            if matchNext:
                message = output[text_start:matchNext.start()]
            else:
                message = output[text_start:]

            log_items.append( {
                'revision': match.group('revision'),
                'author': match.group('author'),
                'date': match.group('date'),
                'message': self._formatLogMessage(message)
                } )
            match = matchNext

        result['log'] = log_items
        return result

    def pull(self, files, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "pull"]
        if options:
            argv += options.split(" ")
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    def remove(self, files, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "rm"]
        if options:
            argv += options.split(" ")
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    def revert(self, files, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "checkout"]
        if options:
            argv += options.split(" ")
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    def status(self, repobasedir, cwd, relpaths=None, options=None, env=None,
               recursive=False, onlywantmodified=False):
        """Check the status of files in the repository
        @param repobasedir {str} Base directory for the repository.
        @param cwd {str} The base directory for the given relpaths.
        @param relpaths {list} Optional - The relative file paths from cwd that
                               are to be checked.
        @param options {str} Optional - Options to add to status command.
        @param env {dict} Optional - Environment variables.
        @param recursive {bool} Optional - whether to recurse into subdirs.
        @param onlywantmodified {bool} Optional - when true, only returns with
                                   files that have been changed.
        """
        files = {}
        result = {
            'cwd': cwd,
            'files': files
        }

        if self._versionInRange(maxVersion="1.6.99"):
            # git < 1.7.0 does not have status --porcelain; use the old way,
            # which is 1) deprecated, 2) slower
            log.debug("git status: version %r too old, using deprecated codepath",
                      self._version)
            return self._status_1_6(repobasedir, cwd, relpaths, options, env,
                                    recursive, onlywantmodified, result)

        # Unfortunately, status by itself does not give us enough information,
        # so we get git to tell us the files that are different, and we assume
        # all other files are under git control as unmodified.

        argv = [self.executable, "status", "--porcelain"]

        if recursive:
            argv.append("--untracked-files=all")
        else:
            argv.append("--untracked-files=normal")

        if self._versionInRange(minVersion="1.7.2"):
            # --ignored was added in git 1.7.2; earlier versions means we'll not
            # be able to figure out which files are ignored (and treat them as
            # committed) :|
            argv.append("--ignored")

        if options:
            argv += options.split(" ")
        if not relpaths:
            relpaths = ["."]
        argv += ["--"] + relpaths
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        result['argv'] = argv

        from os.path import join, exists, isdir, isfile
        is_windows = sys.platform.startswith("win")

        # Parse and overwrite the files that are different.
        for line in output.splitlines(0):
            if len(line) < 4:
                # Not enough info in this line.
                continue
            repo_relpath = line[3:]
            if is_windows:
                # Convert to the Windows path format.
                repo_relpath = repo_relpath.replace("/", "\\")
            repo_relpath = repo_relpath.rstrip(os.sep)
            fullpath = join(repobasedir, repo_relpath)
            cwd_relpath = fullpath[len(cwd):].lstrip(os.sep)
            assert fullpath.lower().startswith(cwd.lower())
            if not recursive and os.sep in cwd_relpath:
                # This file resides in another directory.
                continue
            status1, status2 = line[0:2]
            if status1 == 'A':
                files[fullpath] = { 'status': 'A',  # Added
                                    'relpath': cwd_relpath }
            elif status1 == 'D':
                files[fullpath] = { 'status': 'D',  # Deleted
                                    'relpath': cwd_relpath }
            elif status2 == 'M':
                files[fullpath] = { 'status': 'C',  # Changed
                                    'relpath': cwd_relpath }
            elif status1 in '?!' or status2 in '?!':
                log.debug('fullpath: %r', fullpath)
                files[fullpath] = { 'status': '?',  # Unknown or Ignored
                                    'relpath': cwd_relpath }
        else:
            # We found no results; this can mean no changes in this directory,
            # _or_ this directory is being ignored.  Ask about the cwd.
            parent, leaf = os.path.split(cwd)
            if parent.startswith(repobasedir): # check that we didn't go outside
                argv = [self.executable, "ls-files", "-t", "--directory",
                        "--others", "--", leaf]
                output, error = self._runCommand(argv, cwd=parent, env=env)
                if output.startswith("? "):
                    # Yep, this directory is being ignored; mark this whole
                    # subtree as ignored and return.  This prevents us from
                    # marking it all as unmodified instead (below).
                    files[cwd] = { 'status': '?',
                                   'relpath': '.' }
                    return result

        # Assume all other files are under scc control and are unmodified. Uses
        # os.listdir, or os.walk to discover the unmodified files.
        if not error:
            for relpath in relpaths:
                path = join(cwd, relpath)
                if path in files:
                    # Already have status for this file.
                    continue
                if isdir(path):
                    if onlywantmodified:
                        # Don't recurse into directories, as we already have
                        # all the changed file information from the status
                        # command.
                        continue
                    if recursive:
                        for dirpath, dirnames, filenames in os.walk(cwd):
                            # Filter out directories that are not under scc control.
                            for dname in dirnames[:]: # a copy so we can modify in place
                                fullpath = join(dirpath, dname)
                                info = files.get(fullpath)
                                if info and info.get('status') == '?':
                                    # Ignored or not part of the repo - filter out.
                                    dirnames.remove(dname)
                            for fname in filenames:
                                fullpath = join(dirpath, fname)
                                if fullpath in files:
                                    # Already have status for this file.
                                    continue
                                cwd_relpath = fullpath[len(cwd):].lstrip(os.sep)
                                files[fullpath] = { 'status': 'H', # Under scc control.
                                                    'relpath': cwd_relpath }
                    else:
                        # Just collect the files in the current working directory.
                        for f in os.listdir(cwd):
                            fullpath = join(cwd, f)
                            if fullpath in files:
                                # Already have status for this file.
                                continue
                            if isfile(fullpath):
                                cwd_relpath = fullpath[len(cwd):].lstrip(os.sep)
                                files[fullpath] = { 'status': 'H', # Under scc control.
                                                    'relpath': cwd_relpath }
                elif exists(path):
                    # Files.
                    files[path] = { 'status': 'H', # Under scc control.
                                    'relpath': relpath }

        return result

    def _status_1_6(self, repobasedir, cwd, relpaths, options,
                    env, recursive, onlywantmodified, result):
        """Git status compatiblity for git >= 1.6 < 1.7.0
        The parameters are the same as status() (plus the return value).
        """
        files = result['files']

        argv = [self.executable, "ls-files", "-t", "--directory"]
        if options is None:
            argv += ["--modified", "--deleted", "--cached", "--others", "--killed"]
        else:
            argv += options.split(" ")
        argv += ["--"] + (relpaths or ["."]) # given paths, or cwd
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error

        for line in output.splitlines():
            try:
                status, relpath = line.split(" ", 1)
            except ValueError:
                continue # no space in line (unpack error)
            if len(status) == 1 and status in "HSMRCK?":
                if not recursive and "/" in relpath:
                    # filter out sub-directories - git doesn't have a
                    # non-recursive mode
                    continue
                fullpath = os.path.join(cwd, relpath)
                log.debug("status: ls-files says %r is %r",
                          fullpath, status)
                files[fullpath] = { 'status': status,
                                    'relpath': relpath }

        # unfortunately, ls-files by itself does not give us enough information,
        # as it does not distinguish between added files and modified files.  We
        # use a second call to the status command to obtain the extra details.
        argv = [self.executable, "status", "--untracked-files=normal"]
        argv += ["--"] + (relpaths or ["."])
        # old versions of git will return an error if nothing needs to be
        # committed (wtf?), so ignore the exit status here
        output, error = self._runCommand(argv, cwd=cwd, env=env,
                                         ignore_exit_status=True)
        result['stdout'] += output
        result['stderr'] += error

        # cheapo enum for stages
        STAGED, UNSTAGED, UNTRACKED = object(), object(), object()

        section = None
        for line in output.splitlines():
            if not line.startswith("#"):
                continue
            if line.startswith("# Changes to be committed:"):
                section = STAGED
            elif line.startswith("# Changed but not updated:"):
                section = UNSTAGED
            elif line.startswith("# Untracked files:"):
                section = UNTRACKED

            if not line.startswith("#\t"):
                continue # not a file

            relpath = line.split("\t", 1)[1] # path relative to cwd
            if relpath.startswith("new file:   ") and section is STAGED:
                relpath = relpath.split(None, 1)[1]
                status = 'A' # added
            elif relpath.startswith("deleted:    ") and section is STAGED:
                relpath = relpath.split(None, 1)[1]
                status = 'D' # deleted
            elif relpath.startswith("modified:   ") and section is UNSTAGED:
                relpath = relpath.split(None, 1)[1]
                status = 'C' # changed
            elif section is UNTRACKED:
                status = '?' # unknown
            else:
                continue # not interested in this file

            if not recursive and "/" in relpath:
                continue # this file is in another directory; skip it
            if relpath.startswith("../"):
                continue # this file is outside the working directory

            fullpath = os.path.join(cwd, relpath.replace("/", os.sep))
            log.debug("status: status says %r is %r", fullpath, status)
            files[fullpath] = { 'status': status,
                                'relpath': relpath }

        return result

    @property
    def _version(self):
        """Get the version string of the git binary
        @note This will cache the result"""

        try:
            mtime = os.stat(self.executable).st_mtime
        except:
            # failed to stat() the executable; not usable
            return None

        if hasattr(self, "_git_version_cache"):
            cached_executable, cached_mtime, version = self._git_version_cache
        else:
            cached_executable, cached_mtime, version = None, None, None

        if self.executable == cached_executable and mtime == cached_mtime:
            # we have a good cached value
            return version

        output, error = self._runCommand([self.executable, "--version"])
        if output.startswith("git version "):
            version = output.strip().split(" ", 2)[2]
        else:
            version = None

        # update the cache
        setattr(self, "_git_version_cache", (self.executable, mtime, version))
        return version
    
    def _versionInRange(self, minVersion=None, maxVersion=None):
        """Check that the git executable is within some version range
        @param minVersion {str} The minimum git version; None means no minimum
        @param maxVersion {str} The maximum git version; None means no maximum
        @returns {bool} True if the git executable is within the specified
            version range (inclusive), False if not.
        """
        from distutils.version import LooseVersion
        version = self._version
        if version is None:
            # no version available
            return False
        version = LooseVersion(version)
        if minVersion is not None and version < LooseVersion(minVersion):
            return False
        if maxVersion is not None and version > LooseVersion(maxVersion):
            return False
        return True
